/* The Lag, demo for Linux + GGI
   Copyright (C) 1998 Tijs van Bakel aka Smoke of CRAP
   
The Lag is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

The Lag is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with The Lag; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

/*

   Email me with whatever comments on this at smoke@casema.net
   If you think you can do better; apply for a great position at CRAP.

*/


/*** Define ROOT in order to have a decent framerate.
     When will a vertical-retrace handler be implemented in GGI ? ***/
#undef ROOT

#include <ggi/ggi.h>
#include <gbm.h>

#include <stdio.h>
#include <unistd.h>
#include <math.h>
#include <fcntl.h>

#ifdef __GLIBC2__
	#include <sys/io.h>
	#include <sys/perm.h>
#endif

#define VWIDTH	320
#define VHEIGHT	200
#define VDEPTH GT_8BIT

/* guy with the red head */
#define SPRITE1W 140
#define SPRITE1H 200
#define SPRITE1_FILENAME	"sprite1.gif"

/* logo */
#define SPRITE2W 300
#define SPRITE2H 65
#define SPRITE2_FILENAME	"sprite2.gif"

/* background setting */
#define BACKDROP_FILENAME	"backdrop.gif"

typedef unsigned char pixel;

pixel *VidBuf, *Sprite1, *Sprite2, *Backdrop, *VidBase;
ggi_color ColorMap[256];

#ifdef ROOT
	static __inline__ void port_out(int value, int port)
	{
		__asm__ volatile ("outb %0,%1"
			: : "a" ((unsigned char) value), "d" ((unsigned short) port));
	}

	static __inline__ int port_in(int port)
	{
		unsigned char value;
			__asm__ volatile ("inb %1,%0"
			: "=a" (value) : "d" ((unsigned short) port));
		return value;
	}

	void WaitRetrace(void)
	{
		while (!(port_in(0x3da) & 8))
	        while (port_in(0x3da) & 8)
	}
#endif
 
void Error(char* str)
{
	fprintf(stderr, str);
	exit(1);
}

void LoadPics(void)
{
	GBM *gbmHeader;
	GBMRGB *gbmPalette;
	pixel *gbmData;
	int fd, fileType;

        int x,y,pos,i;

        char *fileNameBD = BACKDROP_FILENAME;
        char *fileNameS1 = SPRITE1_FILENAME;
        char *fileNameS2 = SPRITE2_FILENAME;
        
	gbmHeader = (GBM*) malloc(sizeof(GBM));
	gbmPalette = (GBMRGB*) malloc(256*sizeof(GBMRGB));

                /* load backdrop */
        
	gbm_guess_filetype(fileNameBD, &fileType);
	fd = gbm_io_open(fileNameBD, O_RDONLY);
	gbm_read_header(fileNameBD, fd, fileType, gbmHeader, "");
	gbmData = (pixel*) malloc(gbmHeader->w*gbmHeader->h);
        if (gbmData == NULL)
            Error("cannot free memory for the backdrop picture.");
	gbm_read_palette(fd, fileType, gbmHeader, gbmPalette);
	gbm_read_data(fd, fileType, gbmHeader, gbmData);
	gbm_io_close(fd);

	for (i=0; i<256; i++) {
		ColorMap[i].r = gbmPalette[i].r << 8;
		ColorMap[i].g = gbmPalette[i].g << 8;
		ColorMap[i].b = gbmPalette[i].b << 8;
	}
        
	Backdrop = (pixel*) malloc(320*200);
        pos = 640*480;
	for (y=0;y<200-1;y++) {
            pos -= 320;
            for (x=0;x<320;x++) {
		Backdrop[319-x+y*320] = gbmData[pos--];
            }
        }

        free(gbmData);

                /* load guy */

        Sprite1 = (pixel*) malloc(SPRITE1W*SPRITE1H);
	gbm_guess_filetype(fileNameS1, &fileType);
	fd = gbm_io_open(fileNameS1, O_RDONLY);
	gbm_read_header(fileNameS1, fd, fileType, gbmHeader, "");
	gbmData = (pixel*) malloc(gbmHeader->w*gbmHeader->h);
        if (gbmData == NULL)
            Error("cannot free memory for the guy picture.");
	gbm_read_data(fd, fileType, gbmHeader, gbmData);
	gbm_io_close(fd);
        pos = 640*480;
	for (y=0;y<SPRITE1H-1;y++) {
            pos -= (640-SPRITE1W);
            for (x=0;x<SPRITE1W;x++) {
		Sprite1[SPRITE1W-x+y*SPRITE1W] = gbmData[pos--];
            }
        }
        free(gbmData);

                /* load logo */
        
        Sprite2 = (pixel*) malloc(SPRITE2H*SPRITE2W);
	fd = gbm_io_open(fileNameS2, O_RDONLY);
	gbm_read_header(fileNameS2, fd, fileType, gbmHeader, "");
	gbmData = (pixel*) malloc(gbmHeader->w*gbmHeader->h*2);
        if (gbmData == NULL)
            Error("cannot free memory for the logo.");
	gbm_read_data(fd, fileType, gbmHeader, gbmData);
	gbm_io_close(fd);
        pos = 640*480;
	for (y=0;y<SPRITE2H-1;y++) {
            pos -= (640-SPRITE2W);
            for (x=0;x<SPRITE2W;x++) {
		Sprite2[SPRITE2W-x+y*SPRITE2W] = gbmData[pos--];
            }
        }

        free(gbmHeader);
        free(gbmPalette);
        free(gbmData);
}

void DrawBackdrop(void)
{
    memcpy(VidBuf, Backdrop, 64000);
}

void DrawSprite1(int x, int y)
{
    int i,j,pos,spos;
    pixel k;
    
    for (i = 0; i < SPRITE1H-1; i++)
    {
        if ( ( (y+i) > 0) && ( (y+i) < 200) ) {
            pos = (y+i) * VWIDTH + x;
            spos = i * SPRITE1W +1;
               
            for (j = 0; j < SPRITE1W; j++) {
                k = Sprite1[spos++];
                if (k > 0) 
                    VidBuf[pos+j] = k;
            }
        }
    }
}
void DrawSprite2(int x, int y)
{
    int i,j,pos,spos;
    pixel k;
    
    for (i = 0; i < SPRITE2W-1; i++)
    {
        if ( ( (x+i) > 0) && ( (x+i) < 320) ) {
            pos = y * VWIDTH + i + x;
            spos = i+1;
               
            for (j = 0; j < SPRITE2H-2; j++) {
                k = Sprite2[spos+=SPRITE2W];
                pos+=VWIDTH;
                if (k > 0) 
                    VidBuf[pos] = k;
            }
        }
    }
}

void MainLoop(int frame)
{
    int y,x,y2;
    float f;
    
    VidBuf[frame] = frame;
    DrawBackdrop();

    f = sin( (float) frame / 10 );
    y2 = f * 20;
    f *= f * 21;
    y = f + 20;

    f = cos( (float) frame / 100 );
    f *= 360;
    x = f + 20;
    
    DrawSprite1(130,y);
    DrawSprite2(x,100+y2);

}

int main(void)
{
	int frameCount = 0,
            page = 0;

        ggi_visual_t visual;
	ggi_directbuffer *db;

#ifdef ROOT
	if ((ioperm(0x03DA,1,1)) || (ioperm(0x03C8,1,1)) || (ioperm(0x03C9,1,1))) 
		Error("please hack your way to root, and try again.");
#endif

	if (ggiInit() != 0)
		Error("cannot init libggi.");
	if ((visual = ggiOpen(NULL)) == NULL)
		Error("cannot open default visual.");
	if (ggiSetGraphMode(visual, VWIDTH,VHEIGHT,VWIDTH,VHEIGHT*2,
	                        VDEPTH) != 0)
		Error("cannot set defaultmode");

/*** cannot get directbuffer to work ***/

/*	if (db = ggiDBGetBuffer(visual,0)) {
		error("ggiDBGetBuffer");
	}
	if (db->layout != blPixelLinearBuffer) {
		error("ggiDBGetLayout");
	}
	VidBase = (unsigned char*) db->write; */
	
	VidBuf = (pixel*) malloc(VWIDTH*VHEIGHT*sizeof(pixel));

	LoadPics();
	ggiSetPalette(visual,0,256,ColorMap);

	while (!ggiKbhit(visual)) {

/*** what is the proper way to do page-flipping in ggi ? ***/

/*          if (page == 1) {
                page = 0;
                ggiSetOrigin(visual,0,0);
                VidBuf = VidBase + 320*200;
            }
            else {
                page = 1;
                ggiSetOrigin(visual,0,200);
                VidBuf = VidBase;
            } */

            MainLoop(frameCount++);

#ifdef ROOT
/*** use this with SVGALIB ***/
            WaitRetrace();
#else
	    usleep(10);
#endif

	    ggiPutBox(visual,0,0,320,200,VidBuf);
	}
	
	ggiClose(visual);

        return (0);
}

